import numpy as np
#np.random.seed(592)

import tensorflow as tf
import tensorflow_probability as tfp
tfd = tfp.distributions
#tf.set_random_seed(2)

import numpy as np
import model_viz as mv
import matplotlib.pyplot as plt

import warnings
warnings.filterwarnings("ignore")

# ---------------------------------------------------------------------------------------------------------------------
# Base Function
# ---------------------------------------------------------------------------------------------------------------------

def build(self, bayes):

    x = tf.placeholder(shape=(None, self.num_features), dtype=tf.float32, name="x")
    y = tf.placeholder(shape=(None, self.num_outputs), dtype=tf.float32, name="y")

    d = x
    for layer in range(self.layers):
        layer +=1

        l = tf.layers.dense(inputs=d,
                            units=self.units,
                            activation=self.activation,
                            kernel_regularizer=self.reg,
                            name="layer_{}".format(layer),
                            reuse=tf.AUTO_REUSE)

        d = tf.layers.dropout(inputs=l,
                              rate=self.dropout_rate,
                              noise_shape = (1, self.units),
                              training=bayes,
                              name="dropout_{}".format(layer))

        return x, y, d

# ---------------------------------------------------------------------------------------------------------------------
# MLP
# ---------------------------------------------------------------------------------------------------------------------

def build_mlp(self):

    x, y, d = build(self, False)

    mu = tf.layers.dense(inputs=d,
                            units=self.num_outputs,
                            activation=None,
                            name="mu",
                            reuse=tf.AUTO_REUSE)

    loss = tf.losses.mean_squared_error(y, mu)
    if self.reg is not None:
        loss += tf.losses.get_regularization_losses()

    return {
        "x" : x,
        "y" : y,
        "mu" : mu,
        "loss" : loss
    }

# ---------------------------------------------------------------------------------------------------------------------
# BNDropout
# ---------------------------------------------------------------------------------------------------------------------

def build_bn_dropout(self):

    if self.lengthscale is not None:
        tau = 1.0
        N = self.train_samples
        reg = self.lengthscale**2 * (1 - self.dropout_rate) / (2. * N * tau)
        self.reg = tf.contrib.layers.l2_regularizer(reg)

    x, y, d = build(self, True)

    mu = tf.layers.dense(inputs=d,
                            units=self.num_outputs,
                            activation=None,
                            kernel_regularizer=self.reg,
                            name="mu",
                            reuse=tf.AUTO_REUSE)

    sigma = tf.layers.dense(inputs=d,
                            units=self.num_outputs,
                            activation=tf.nn.softplus,
                            kernel_regularizer=self.reg,
                            name="sigma",
                            reuse=tf.AUTO_REUSE)

    likelihood = tfd.Normal(loc=mu, scale=sigma)

    loss = tf.reduce_mean(-likelihood.log_prob(y))
    if self.reg is not None:
        loss += tf.losses.get_regularization_losses()

    return {
        "x" : x,
        "y" : y,
        "mu" : mu,
        "sigma" : sigma,
        "loss" : loss
    }

# ---------------------------------------------------------------------------------------------------------------------
# BNVI
# ---------------------------------------------------------------------------------------------------------------------

def VariationalParameter(name, shape, constraint=None, mean=0.0, std=1.0):
    means = tf.get_variable(name+'_mean', initializer = mean*tf.ones(shape), constraint=constraint)
    stds = tf.nn.softplus( tf.get_variable(name+'_std', initializer = std*tf.ones(shape)) )
    return tfd.Normal(loc=means, scale=stds)

def build_layer(self, input_tensor, w_shape, b_shape, layer_num, activation=None):

        # Prior
        pw = tfd.Normal(name='pw{}'.format(layer_num),
                        loc=self.p_mean*tf.ones(w_shape),
                        scale=self.p_std*tf.ones(w_shape))
        pb = tfd.Normal(name='pb{}'.format(layer_num),
                        loc=self.p_mean*tf.ones(b_shape),
                        scale=self.p_std*tf.ones(b_shape))

        # Variational Parameters
        qw = VariationalParameter('w{}'.format(layer_num), w_shape,
                                  mean=self.q_mean,
                                  std=self.q_std)
        qb = VariationalParameter('b{}'.format(layer_num), b_shape,
                                  mean=self.q_mean,
                                  std=self.q_std)

        # KL divergence with priors
        kl_w = tfp.distributions.kl_divergence(qw, pw,
                                               allow_nan_stats=True,
                                               name="kl_w{}".format(layer_num))
        kl_b = tfp.distributions.kl_divergence(qb, pb,
                                               allow_nan_stats=True,
                                               name="kl_b{}".format(layer_num))
        kl = tf.math.reduce_sum(kl_w) + tf.math.reduce_sum(kl_b)


        h =  tf.matmul(input_tensor, qw.sample()) + qb.sample()

        if activation is not None:
            h =  activation( h )

        return h, kl

def build_bn_vi(self):
    # Placeholders
    x = tf.placeholder(tf.float32, [None, self.num_features], name="x")
    y = tf.placeholder(tf.float32, [None, self.num_outputs], name="y")

    KL = []

    h, kl = build_layer(self, x, [self.num_features, self.units], [self.units], 0, self.activation)
    KL.append(kl)

    layer = -1
    for layer in range(self.layers):
        h, kl = build_layer(self, h, [self.units, self.units], [self.units], layer+1, self.activation)
        KL.append(kl)

    mu, kl = build_layer(self, h, [self.units, self.num_outputs], [self.num_outputs], layer+2)
    KL.append(kl)

    # Heteroskedastic
    sigma, kl = build_layer(self, h, [self.units, self.num_outputs], [self.num_outputs], "std", tf.nn.softplus)
    KL.append(kl)
    # Posterior Predictive
    pred_distribution = tfd.Normal(loc=mu, scale=sigma+1e-5)

    # Homoscedastic
    #sigma = VariationalParameter('std_', [self.num_outputs], mean=self.q_mean, std=self.q_std)
    # Posterior Predictive
    #pred_distribution = tfd.Normal(loc=mu, scale=tf.nn.softplus(sigma.sample()))

    # Probability of train data
    neg_log_prob = tf.reduce_mean(-pred_distribution.log_prob(y))

    prior_loss = 0.0
    for kl in KL:
        prior_loss += kl

    elbo_loss = neg_log_prob + prior_loss / self.train_samples

    return {
        'x' : x,
        'y' : y,
        'mu' : mu,
        'sigma' : sigma,
        'loss' : elbo_loss
    }



import random
import time

def shuffle(x, y, N):
    ind_list = [i for i in range(N)]
    random.shuffle(ind_list)
    x = x[ind_list]
    y = y[ind_list]
    return x, y

class Timer:
    def __enter__(self):
        self.t0 = time.time()
    def __exit__(self, *args):
        print('Elapsed time: %0.2fs' % (time.time()-self.t0))



import setuptools

# python3 -m pip install --user --upgrade setuptools wheel
# python3 setup.py sdist bdist_wheel
# python3 -m twine upload --repository-url https://upload.pypi.org/legacy/ dist/*
# python3 -m pip install kassandra

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kassandra",
    version="0.0.12",
    author="James Montgomery",
    author_email="jamesoneillmontgomery@gmail.com",
    description="Package for implementing bayesian deep learning models in python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/James-Montgomery/kassandra",
    packages=setuptools.find_packages(),
    install_requires=[
        'tensorflow==1.13.1',
        'tensorflow-probability==0.6.0',
        'numpy>=1.15.1',
        'scipy>=1.1.0',
        'matplotlib>=2.2.3',
        'seaborn>=0.9.0',
        'nose>=1.3.7',
        'coverage==4.0.0',
        'pylint>=1.8.2'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import scipy.stats as st
import numpy as np
#np.random.seed(592)

import tensorflow as tf
import progressbar
#tf.set_random_seed(2)

from . import build_utils
from . import utils

import warnings
warnings.filterwarnings("ignore")

# ---------------------------------------------------------------------------------------------------------------------
# Base Class
# ---------------------------------------------------------------------------------------------------------------------

class NeuralNetwork(object):
    """Neural Network Base Class"""

    def __init__(self, num_features=1, num_outputs=1, layers=2, units=50, activation=tf.nn.relu, optimizer=None):

        # inputs and outputs
        self.num_features = num_features
        self.num_outputs = num_outputs

        # build specifications
        self.layers = layers
        self.activation = activation
        self.units = units
        self.reg = None

        # set up the graph
        tf.reset_default_graph()
        self.tensors = self._build()
        self.session = tf.Session()

        # training specifications
        if optimizer is None:
            optimizer = tf.train.AdamOptimizer(learning_rate=0.01)
        self.step_op = optimizer.minimize(self.tensors["loss"])

        # start session
        self.session.run(tf.global_variables_initializer())

    def train(self, x_train, y_train, epochs=100, show_progress=True, show_loss=False, shuffle=False, minibatches=1):

        N = x_train.shape[0]

        if shuffle:
            x_train, y_train = utils.shuffle(x_train, y_train, N)

        # set up progress bar
        if show_progress:
            pbar = progressbar.ProgressBar()
            n_iter = pbar(range(epochs))
        else:
            n_iter = range(epochs)

        batch_size = int(N / minibatches)
        assert batch_size > 0, "Specified too many minibatches for number of train samples."

        # start training
        with utils.Timer():
            for i in n_iter:

                batch_idx = np.random.choice(N, batch_size, replace=False)

                feed_dict = {self.tensors["x"] : x_train[batch_idx].reshape(-1, self.num_features),
                            self.tensors["y"] : y_train[batch_idx].reshape(-1, self.num_outputs)}

                l, _ = self.session.run([self.tensors["loss"], self.step_op], feed_dict)


                if show_loss and i % 1000 ==0:
                    print("loss {}".format(l))

# ---------------------------------------------------------------------------------------------------------------------
# MLP
# ---------------------------------------------------------------------------------------------------------------------

class MLP(NeuralNetwork):
    """Multi-Layer Perceptron"""

    def __init__(self, dropout_rate=0.5, *args, **kwargs):
        self.name = "MLP"
        self.dropout_rate = dropout_rate
        NeuralNetwork.__init__(self, *args, **kwargs)

    def _build(self):
        return build_utils.build_mlp(self)

    def predict(self, x_, n_estimates=10):
        feed_dict = {self.tensors["x"] : x_[:, np.newaxis]}
        return np.asarray(self.session.run(self.tensors["mu"], feed_dict) ).squeeze()

# ---------------------------------------------------------------------------------------------------------------------
# BNDropout
# ---------------------------------------------------------------------------------------------------------------------

class BNDropout(NeuralNetwork):
    """Bayesian Neural Network using MC Dropout"""

    def __init__(self, dropout_rate=0.5, train_samples=100, lengthscale=None, *args, **kwargs):
        self.name = "BNDropout"
        self.dropout_rate = dropout_rate
        self.train_samples = train_samples
        self.lengthscale = lengthscale

        NeuralNetwork.__init__(self, *args, **kwargs)

    def _build(self):
        return build_utils.build_bn_dropout(self)

    def predict(self, x_test, n_estimates=10):
        feed_dict = {self.tensors["x"] : x_test[:, np.newaxis]}

        samples = np.asarray([
            self.session.run([self.tensors["mu"], self.tensors["sigma"]], feed_dict)
            for i in range(n_estimates)
        ]).squeeze()

        mu, sigma = samples[:,0,:].mean(axis=0), samples[:,1,:].mean(axis=0)

        return np.asarray([mu, sigma]).squeeze()

    def sample_posterior(self, x_test):
        feed_dict = {self.tensors["x"] : x_test[:, np.newaxis]}
        posterior_sample = self.session.run([self.tensors["mu"], self.tensors["sigma"]], feed_dict)
        return np.asarray(posterior_sample).squeeze()

    def sample_posterior_predictive(self, x_test):
        posterior_sample = self.sample_posterior(x_test)
        mean, cov = posterior_sample[0,:], np.eye(posterior_sample.shape[1]) * posterior_sample[1,:]
        return st.multivariate_normal(mean=mean, cov=cov).rvs(1)

# ---------------------------------------------------------------------------------------------------------------------
# BNVI
# ---------------------------------------------------------------------------------------------------------------------

class BNVI(NeuralNetwork):
    """Bayesian Neural Network using Variational Inference"""

    def __init__(self, train_samples=100, p_mean=0.0, p_std=1.0, q_mean=0.0, q_std=1.0, *args, **kwargs):
        self.name = "BNVI"
        self.train_samples = train_samples

        # prior parameters
        self.p_mean = p_mean
        self.p_std = p_std

        # variational parameters
        self.q_mean = q_mean
        self.q_std = q_std

        NeuralNetwork.__init__(self, *args, **kwargs)

    def _build(self):
        return build_utils.build_bn_vi(self)

    def predict(self, x_test, n_estimates=10):
        feed_dict = {self.tensors["x"] : x_test[:, np.newaxis]}

        samples = np.asarray([
            self.session.run([self.tensors["mu"], self.tensors["sigma"]], feed_dict)
            for i in range(n_estimates)
        ]).squeeze()

        mu, sigma = samples[:,0,:].mean(axis=0), samples[:,1,:].mean(axis=0)

        return np.asarray([mu, sigma]).squeeze()

    def sample_posterior(self, x_test):
        feed_dict = {self.tensors["x"] : x_test[:, np.newaxis]}
        posterior_sample = self.session.run([self.tensors["mu"], self.tensors["sigma"]], feed_dict)
        return np.asarray(posterior_sample).squeeze()

    def sample_posterior_predictive(self, x_test):
        posterior_sample = self.sample_posterior(x_test)
        mean, cov = posterior_sample[0,:], np.eye(posterior_sample.shape[1]) * posterior_sample[1,:]
        return st.multivariate_normal(mean=mean, cov=cov).rvs(1)
